package remoteapi.server;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.nio.ByteBuffer;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.google.appengine.api.datastore.DatastoreService;
import com.google.appengine.api.datastore.DatastoreServiceFactory;
import com.google.apphosting.api.ApiProxy;
import com.google.apphosting.api.DatastorePb;
import com.google.apphosting.api.ApiProxy.ApplicationException;
import com.google.apphosting.api.DatastorePb.GetSchemaRequest;
import com.google.apphosting.api.DatastorePb.Query;
import com.google.apphosting.api.DatastorePb.QueryResult;
import com.google.apphosting.api.DatastorePb.Schema;
import com.google.storage.onestore.v3.OnestoreEntity;
import com.google.storage.onestore.v3.OnestoreEntity.EntityProto;
import com.google.storage.onestore.v3.OnestoreEntity.Property;
import com.google.storage.onestore.v3.OnestoreEntity.PropertyValue;
import com.google.storage.onestore.v3.OnestoreEntity.Path.Element;

public class MakeSyncCallServlet extends HttpServlet {

	/**
	 * リクエストされたbyte配列を{@link ApiProxy#makeSyncCall(String, String, byte[])}へ引き渡し、
	 * その結果をレスポンスするだけのServlet.
	 * <p>特殊なサーブレットなので、web.xmlにてセキュリティを設定しておくのがおすすめ。</p>
	 * <div><ul>
	 * <li>HttpHeaderに以下を設定する。
	 * <ul><li>{@literal serviceName}</li><li>{@literal methodName}</li>
	 * <li>{@litera applicationId}</li></ul></li>
	 * <li>payloadにProtocolBufferで出力されたbyte配列を設定して{@literal POST}する。</li>
	 * <li>{@literal application/octet-stream}で
	 * {@link ApiProxy#makeSyncCall(String, String, byte[])}の結果を返すので、クライアント側でよしなに。</li>
	 * </ul></div>
	 * 
	 * @author shin1ogawa
	 */
	
	/*
	 * このソースは変更されています。オリジナルは下記URLを参照。
	 * http://shin1o.blogspot.com/2009/10/appengine-makesynccallservlet.html
	 * */
	
	  private static final long serialVersionUID = 2380791176214953417L;
	  private static final String APPLICATION_ID = "applicationId";
	  private static final String METHOD_NAME = "methodName";
	  private static final String SERVICE_NAME = "serviceName";
	  private static Logger logger = Logger.getLogger(MakeSyncCallServlet.class.getName());

	protected void doGet(HttpServletRequest req, HttpServletResponse resp)
			throws ServletException, IOException {
		
			resp.setStatus(HttpURLConnection.HTTP_OK);
			
	}
	
	@Override
	protected void doPost(HttpServletRequest req, HttpServletResponse resp)
		throws ServletException, IOException {
		
		String serviceName = req.getHeader(SERVICE_NAME);
		String methodName = req.getHeader(METHOD_NAME);
		String applicationId = req.getHeader(APPLICATION_ID);
		logger.info(applicationId + ":" + serviceName + "#" + methodName);
		if (validateParameters(resp, serviceName, methodName, applicationId) == false) {
			return;
		}
		byte[] requestBytes = toByteArray(req.getInputStream());

		logger.info(applicationId + ":" + serviceName + "#" + methodName + ": requestBytes.length="
		  + (requestBytes != null ? requestBytes.length : "null"));
		
		String environment = System.getProperty("com.google.appengine.runtime.environment");
		
		if(environment.equals("Development")){
			
			Query query = new DatastorePb.Query();
			query.parseFrom(requestBytes);
			
			if(query.getKind().equals("__Stat_Kind__")){
		
				Schema schema = getSchema();

				byte[] bytes = ApiProxy.makeSyncCall(serviceName, methodName, requestBytes);
			    QueryResult qr = new QueryResult();
			    qr.parseFrom(bytes);
				
				byte[] responseBytes;
				if(qr.hasCursor()){
					responseBytes = statKindQueryResultEmu(schema,query.getOffset(),query.getLimit());
				}else{
					responseBytes = statKindCountEntitiesQueryResultEmu(schema.kinds().size());
				}
				
				resp.setContentType("application/octet-stream");
				resp.addHeader("Content-Length", String.valueOf(responseBytes.length));
				resp.getOutputStream().write(responseBytes);
				resp.getOutputStream().flush();
				return;
			}
		}

		byte[] responseBytes = null;
		try {
			@SuppressWarnings("unchecked")
			byte[] bytes = ApiProxy.getDelegate().makeSyncCall(ApiProxy.getCurrentEnvironment(),
		    serviceName, methodName, requestBytes);
			responseBytes = bytes;
		} catch (ApplicationException ae) {
			logger.log(Level.WARNING, applicationId + ":" + serviceName + "#" + methodName, ae);
			resp.setStatus(HttpURLConnection.HTTP_INTERNAL_ERROR);
			resp.setContentType("text/plain");
			ae.printStackTrace(resp.getWriter());
			resp.getWriter().flush();
			return;
		} catch (Throwable th) {
			logger.log(Level.WARNING, applicationId + ":" + serviceName + "#" + methodName, th);
			resp.setStatus(HttpURLConnection.HTTP_INTERNAL_ERROR);
			resp.setContentType("text/plain");
			th.printStackTrace(resp.getWriter());
			resp.getWriter().flush();
			return;
		}
		if (responseBytes == null) {
			logger.info(serviceName + "#" + methodName + ": responseBytes == null.");
			responseBytes = new byte[0];
		} else {
			logger.info(serviceName + "#" + methodName + ": responseBytes.length="
					+ responseBytes.length);
		}
		resp.setContentType("application/octet-stream");
		resp.addHeader("Content-Length", String.valueOf(responseBytes.length));
		resp.getOutputStream().write(responseBytes);
		resp.getOutputStream().flush();
	}

	Schema getSchema(){
		GetSchemaRequest schemaRequest = new GetSchemaRequest();
		schemaRequest.setApp(ApiProxy.getCurrentEnvironment().getAppId());
		@SuppressWarnings("unchecked")
		byte[] responseBytes = ApiProxy.getDelegate().makeSyncCall(ApiProxy.getCurrentEnvironment(),
				"datastore_v3", "getSchema", schemaRequest.toByteArray());
		Schema schema = new Schema();
		schema.mergeFrom(responseBytes);
		return schema;
	}
	
	byte[] statKindQueryResultEmu(Schema schema, int offset, int limit){

		System.out.println(schema);
		QueryResult queryResult = new DatastorePb.QueryResult();
	
		List<EntityProto> kinds = schema.kinds();
		if(offset<0) 	offset = 0;
		if(limit<0) limit = kinds.size();
		
		EntityProto entityProto;
		Element statKindElm;
		String kind;
		for (int i = offset; i < offset+limit; i++) {
			if(i>=kinds.size())break;
			entityProto = kinds.get(i);
			kind = entityProto.getKey().getPath().getElement(0).getType();
			
			statKindElm = new OnestoreEntity.Path.Element();
			statKindElm.setType("__Stat_Kind__");
			statKindElm.setName(kind);
			
			//set key
			entityProto.getKey().getPath().clearElement();
			entityProto.getKey().setApp(ApiProxy.getCurrentEnvironment().getAppId());
			entityProto.getKey().getPath().addElement(statKindElm);
			entityProto.getEntityGroup().addElement(statKindElm);
	
			//set kindname
			Property kindProperty = entityProto.addProperty();
			kindProperty.setName("kind_name");
			kindProperty.setValue(new PropertyValue().setStringValue(kind));
			
			com.google.appengine.api.datastore.Query countQuery;
			countQuery = new com.google.appengine.api.datastore.Query(kind);
			countQuery.setKeysOnly();
			
			DatastoreService service = DatastoreServiceFactory.getDatastoreService();
			int entitySize = service.prepare(countQuery).countEntities();
			
			//set count
			Property countProperty = entityProto.addProperty();
			countProperty.setName("count");
			countProperty.setValue(new PropertyValue().setInt64Value(entitySize));
			
			queryResult.addResult(entityProto);
		}
		queryResult.setKeysOnly(false);
		queryResult.setMoreResults(false);
		
		return queryResult.toByteArray();
	}
	
	byte[] statKindCountEntitiesQueryResultEmu(int count){
		if(count<256){
			ByteBuffer buffer = ByteBuffer.allocate(2);
			buffer.put((byte)8);
			buffer.put((byte)count);
			return buffer.array();
		}
		if(count<256*256){
			ByteBuffer buffer = ByteBuffer.allocate(3);
			buffer.put((byte)16);
			buffer.putShort((short)count);
			return buffer.array();
		}
		if(count<256*256*256*256){
			ByteBuffer buffer = ByteBuffer.allocate(5);
			buffer.put((byte)32);
			buffer.putInt(count);
			return buffer.array();
		}
		return new byte[0];
	}
	
	private byte[] toByteArray(InputStream is){
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		byte[] buff = new byte[8192];
		int r;
		try {
			while((r = is.read(buff))!=-1){
				baos.write(buff, 0, r);
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
		return baos.toByteArray();
	}
	
	private boolean validateParameters(HttpServletResponse resp, String serviceName,
		String methodName, String applicationId) throws IOException {
		if (serviceName==null||serviceName.equals("")) {
			onErrorInParameters(resp, "serviceName was not specified.");
			return false;
		}
		if (methodName==null||methodName.equals("")) {
			onErrorInParameters(resp, "methodName was not specified.");
			return false;
		}
		if (applicationId==null||applicationId.equals("")) {
			onErrorInParameters(resp, "applicationId was not specified.");
			return false;
		}
		// 念のためサーバ環境のApplicationIdと同じかどうか確認する。
		String serverApplicationId = ApiProxy.getCurrentEnvironment().getAppId();
		if (serverApplicationId.equals(applicationId) == false) {
			onErrorInParameters(resp, "applicationId was not equals to " + serverApplicationId
			    + ".");
			return false;
		}
		return true;
	}

	private void onErrorInParameters(HttpServletResponse resp, String x) throws IOException {
		resp.setStatus(HttpURLConnection.HTTP_INTERNAL_ERROR);
		resp.setContentType("text/plain");
		resp.getWriter().println(x);
		resp.getWriter().flush();
	}
}
